from __future__ import annotations

from pydantic import BaseModel, Field, ConfigDict

from predibase.resources.deployment import Deployment


class BatchInferenceServerConfig(BaseModel):
    model_config = ConfigDict(extra="allow")

    custom_args: list[str] | None = Field(
        default=None,
        serialization_alias="customArgs",
    )
    hf_token: str | None = Field(
        default=None,
        serialization_alias="hfToken",
    )
    lorax_image_tag: str | None = Field(
        default=None,
        serialization_alias="loraxImageTag",
    )
    base_model: str = Field(
        ...,
        serialization_alias="baseModel",
    )
    quantization: str | None = Field(default=None)

    @classmethod
    def from_deployment(cls, deployment: Deployment) -> BatchInferenceServerConfig:
        return BatchInferenceServerConfig(
            base_model=deployment.model,
            quantization=deployment.quantization,
            custom_args=deployment.config.custom_args,
            hf_token=deployment.config.hf_token,
            lorax_image_tag=deployment.config.lorax_image_tag,
        )



from __future__ import annotations

from pydantic import BaseModel, Field


class BatchInferenceJob(BaseModel):
    uuid: str
    original_request: dict = Field(validation_alias="originalRequest")
    status: str
    error: str | None = Field(default=None)
    result_path: str | None = Field(default=None, validation_alias="resultPath")



from predibase.util.metrics import model_logs_directory


def launch_tensorboard(
    model_id: int,
) -> None:
    logs_dir_name = model_logs_directory(model_id)

    try:
        import tensorboard.notebook
    except ImportError:
        print("Tensorboard not installed. Install with `pip install tensorboard`")
        raise ImportError("Tensorboard not installed. Install with `pip install tensorboard`")

    # TODO: Detect if default port is in use and select another one.
    tensorboard.notebook.start(f"--logdir={logs_dir_name}")



import logging
import os

from ._client import Predibase
from .client import PredibaseClient
from .config import (
    AugmentationConfig,
    ContinuedPretrainingConfig,
    DeploymentConfig,
    FinetuningConfig,
    GRPOConfig,
    SamplingParamsConfig,
    SFTConfig,
    UpdateDeploymentConfig,
    RewardFunction,
    RewardFunctionsRuntimeConfig,
    RewardFunctionsConfig,
)
from .predictor import AsyncPredictor, Predictor
from .resource.llm.prompt import PromptTemplate
from .version import __version__  # noqa: F401

logger = logging.getLogger(__name__)

__all__ = [
    "AsyncPredictor",
    "AugmentationConfig",
    "ContinuedPretrainingConfig",
    "DeploymentConfig",
    "FinetuningConfig",
    "GRPOConfig",
    "Predibase",
    "PredibaseClient",
    "Predictor",
    "PromptTemplate",
    "RewardFunction",
    "RewardFunctionsConfig",
    "RewardFunctionsRuntimeConfig",
    "SFTConfig",
    "SamplingParamsConfig",
    "UpdateDeploymentConfig",
]


def _configure_logging():
    log_level = os.getenv("PREDIBASE_LOG_LEVEL", "").lower()
    requests_logger = logging.getLogger("requests")
    urllib3_logger = logging.getLogger("urllib3")

    if not log_level:
        requests_logger.setLevel(logging.CRITICAL)
        urllib3_logger.setLevel(logging.CRITICAL)
    elif log_level == "debug":
        logger.setLevel(logging.DEBUG)
        requests_logger.setLevel(logging.DEBUG)
        urllib3_logger.setLevel(logging.DEBUG)
    elif log_level == "info":
        logger.setLevel(logging.INFO)
        requests_logger.setLevel(logging.INFO)
        urllib3_logger.setLevel(logging.INFO)


_configure_logging()



def remove_prefix(s: str, prefix: str) -> str:
    if not s.startswith(prefix):
        return s

    return s[len(prefix) :]


def remove_suffix(s: str, suffix: str) -> str:
    if not s.endswith(suffix):
        return s

    return s[: -len(suffix)]



from __future__ import annotations

from typing import TYPE_CHECKING

import requests

from predibase.beta.config import BatchInferenceServerConfig
from predibase.beta.resources.batch_inference_job import BatchInferenceJob
from predibase.resources.dataset import Dataset

if TYPE_CHECKING:
    from predibase import Predibase


class BatchInferenceJobs:
    def __init__(self, client: Predibase):
        self._client = client

    def create(
        self,
        *,
        dataset: str | Dataset,
        server_config: BatchInferenceServerConfig,
        # generation_params: dict,  TODO
        # metadata: dict  TODO
    ) -> BatchInferenceJob:

        if isinstance(dataset, Dataset):
            dataset = dataset.name

        resp = self._client.http_post(
            "/v2/batch-inference",
            json={
                "dataset": dataset,
                "serverConfig": server_config.model_dump(by_alias=True),
            },
        )

        job = BatchInferenceJob.model_validate(resp)

        print(
            f"Successfully requested batch inference over {dataset} using {server_config.base_model} as "
            f"{job.uuid}.\n"
        )

        return job

    def get(self, job: str | BatchInferenceJob) -> BatchInferenceJob:
        if isinstance(job, BatchInferenceJob):
            job = job.uuid

        resp = self._client.http_get(f"/v2/batch-inference/{job}")
        return BatchInferenceJob.model_validate(resp)

    def list(self, limit: int = 10) -> list[BatchInferenceJob]:
        resp = self._client.http_get(f"/v2/batch-inference?limit={limit}")
        return [BatchInferenceJob.model_validate(j) for j in resp["data"]["jobs"]]

    def cancel(self, job: str | BatchInferenceJob):
        if isinstance(job, BatchInferenceJob):
            job = job.uuid

        self._client.http_get(f"/v2/batch-inference/{job}/cancel")

    def download_results(self, job: str | BatchInferenceJob, *, dest: str):
        if isinstance(job, BatchInferenceJob):
            job = job.uuid

        resp = self._client.http_get(f"/v2/batch-inference/{job}/results")

        r = requests.get(resp["url"])
        r.raise_for_status()

        with open(dest, "wb") as f:
            f.write(r.content)



from pydantic import AliasPath, BaseModel, Field


class Dataset(BaseModel):
    uuid: str
    name: str
    connection_type: str = Field(validation_alias=AliasPath("connection", "type"))
    connection_name: str = Field(validation_alias=AliasPath("connection", "name"))
    # TODO: re-enable these fields
    # columns: List[str]
    # num_rows: int = Field(validation_alias=AliasPath("datasetInfo", "datasetProfile",
    #    "DatasetProfile", "num_examples"))
    status: str



from typing import Any, Dict, List, Optional, Union

import pandas as pd

from predibase.pql.api import PQLException, Session
from predibase.resource import model as mdl
from predibase.resource.deployment import Deployment
from predibase.util import spinner


class DeploymentMixin:
    session: Session

    @spinner(name="Create Deployment")
    def create_deployment(
        self,
        name: str,
        model: "mdl.Model",
        engine_name: Optional[str] = None,
        replace: bool = False,
        exists_ok: bool = False,
        comment: Optional[str] = None,
    ) -> Union[pd.DataFrame, Deployment]:
        """Creates a deployment.

        :param str name: Name of deployment.
        :param Model model: Model object to deploy.
        :param str engine_name: Optional serving engine name to deploy to (Default None).
        :param bool replace: Optional flag replace a deployment (default False).
        :param bool exists_ok: Optional flag to only create deployment if not exists (default False).
        :param str comment: Optional comment for the deployment (default None).
        :return: pandas DataFrame or list of Deployment objects.
        """
        if self.session.is_free_trial():
            raise PermissionError(
                "Deployments are locked during the trial period. Contact us to upgrade or if you would like a demo",
            )
        elif self.session.is_plan_expired():
            raise PermissionError(
                "Deployments are locked for expired plans. Contact us to upgrade or if you would like a demo",
            )
        else:
            conditions = []
            if replace:
                conditions.append("OR REPLACE")
            conditions.append("DEPLOYMENT")
            if exists_ok:
                conditions.append("IF NOT EXISTS")
            conditions.append(f'"{name}"')

            if engine_name is not None:
                conditions.append(f'TO "{engine_name}"')
            conditions.append(f'USING "{model.repo.name}" VERSION {model.version}')
            if comment is not None:
                conditions.append(f"COMMENT '{comment}'")

            query = f"CREATE {' '.join(conditions)}"
            result = self.session.execute(query)
            deployments = self._format_deployment_result_df(result)

            if len(deployments) == 1:
                return Deployment.from_dict({"session": self.session, **deployments[0]})
            else:
                raise ValueError(
                    "Error creating deployment",
                )

    @spinner(name="Delete Deployment")
    def delete_deployment(
        self,
        name: str,
        if_exists: bool = False,
        df: bool = False,
    ) -> Union[pd.DataFrame, List[Deployment]]:
        """Deletes a deployment.

        :param str name: required name of deployment.
        :param bool if_exists: Optional flag to only delete deployment if it exists (default False).
        :param bool df: Optional flag to return the result as a dataframe (Default False).
        :return: pandas DataFrame or list of Deployment objects.
        """
        conditions = []
        conditions.append("DROP DEPLOYMENT")
        if if_exists:
            conditions.append("IF EXISTS")
        conditions.append(f'"{name}"')

        query = f"{' '.join(conditions)}"
        try:
            result = self.session.execute(query)
            deployments = self._format_deployment_result_df(result)
            if df:
                return pd.DataFrame(deployments)
            return [Deployment.from_dict({"session": self.session, **x}) for x in deployments]
        except PQLException as e:
            print(
                f"Deleting deployment {name} failed. If this was"
                "an LLM delpoyment use the `llm.delete()` method instead.",
            )
            raise e

    def list_deployments(
        self,
        deployment_name_pattern: Optional[str] = None,
        repo_name: Optional[str] = None,
        model_version: Optional[str] = None,
        df: bool = False,
    ) -> Union[pd.DataFrame, List[Deployment]]:
        """Lists (and optionally filters) Deployments.

        :param str deployment_name_pattern: Optional filter for deployment name (Default None).
        :param str engine_name: Optional filter by engine name (Default None).
        :param str repo_name: Optional filter by model repository name (Default None).
        :param str model_version: Optional filter by model version (Default None).
        :param bool df: Optional flag to return the result as a dataframe (Default False).
        :return: pandas DataFrame or list of Deployment objects.
        """
        endpoint = "/deployments"
        deployment_results = self.session.get_json(endpoint)
        assert isinstance(deployment_results, list)

        if deployment_name_pattern:
            deployment_results = [d for d in deployment_results if deployment_name_pattern in d["Name"]]

        if repo_name:
            deployment_results = [d for d in deployment_results if repo_name == d["HeadVersion"]["RepoName"]]

        if model_version:
            deployment_results = [
                d for d in deployment_results if model_version == str(d["HeadVersion"]["ModelVersion"])
            ]

        deployments = self._format_deployment_result(deployment_results)
        if df:
            return pd.DataFrame(deployments)
        return [Deployment.from_dict({"session": self.session, **x}) for x in deployments]

    def _format_deployment_result(self, result: List[Dict]) -> List[Dict[str, Any]]:
        deployments = [
            {
                "session": self.session,
                "name": d["Name"],
                "deploymentUrl": d["URL"],
                "deploymentVersion": d["HeadVersion"]["VersionNumber"],
                "engineName": d["HeadVersion"]["EngineName"],
                "modelName": f'{d["HeadVersion"]["RepoName"]}',
                "modelVersion": str(d["HeadVersion"]["ModelVersion"]),
                "comment": d["HeadVersion"]["Comment"],
                "errorText": d["HeadVersion"]["ErrorText"],
            }
            for d in result
        ]
        return deployments

    def _format_deployment_result_df(self, result: pd.DataFrame) -> List[Dict[str, Any]]:
        deployments = [
            {
                "session": self.session,
                "name": row["name"],
                "deploymentUrl": row["url"],
                "deploymentVersion": row["head_version_number"],
                "engineName": row["head_version_engine_name"],
                "modelName": row["head_version_model_name"],
                "modelVersion": row["head_version_model_version"],
                "comment": row["head_version_comment"],
                "errorText": row["head_version_error_text"],
            }
            for row in result.to_dict(orient="records")
        ]
        return deployments

    def get_deployment(self, name: str) -> Deployment:
        """Gets a deployment.

        :param str name: Name of deployment.
        :return: Deployment object.
        """
        deployments = self.list_deployments()
        if len(deployments) > 0:
            for d in deployments:
                if d.name == name:
                    return d
        raise ValueError(f"Deployment '{name}' not found.")



from __future__ import annotations

from typing import Any

from predibase.config import GRPOConfig, SFTConfig, ContinuedPretrainingConfig, FinetuningConfig
from pydantic import AliasPath, BaseModel, Field, GetCoreSchemaHandler, ConfigDict
from pydantic_core import core_schema, CoreSchema, from_json, to_json


# See https://stackoverflow.com/a/76700794
class FinetuningJobParams(BaseModel):
    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler) -> CoreSchema:

        validate = core_schema.no_info_plain_validator_function(cls.load_as_json)

        return core_schema.json_or_python_schema(
            json_schema=validate,
            python_schema=validate,
            serialization=core_schema.plain_serializer_function_ser_schema(lambda x: to_json(x)),
        )

    @classmethod
    def load_as_json(cls, value: Any):
        if isinstance(value, str):
            return from_json(value)
        return value


class FinetuningJob(BaseModel):
    uuid: str
    base_model: str = Field(validation_alias="baseModel")
    dataset: str = Field(validation_alias="datasetName")
    description: str | None = Field(default=None)
    target_repo: str = Field(validation_alias="targetRepoName")
    target_version_tag: int = Field(validation_alias="targetVersionTag")
    accelerator: str = Field(validation_alias=AliasPath("accelerator", "id"))
    status: str
    params: GRPOConfig | SFTConfig | ContinuedPretrainingConfig | FinetuningConfig | dict
    # duration TODO needs calc


class _GRPOTrainRewards(BaseModel):
    model_config = ConfigDict(extra="allow")

    total_reward: float | None = Field(default=None)
    total_reward_std: float | None = Field(default=None)


class _FinetuningMetricsData(BaseModel):
    batch_size: int | None = Field(default=None)
    best_eval_metric_checkpoint_number: int | None = Field(default=None)
    best_eval_metric_epoch: int | None = Field(default=None)
    best_eval_metric_steps: int | None = Field(default=None)
    best_train_metrics_loss: float | None = Field(default=None)
    best_valid_metric: float | None = Field(default=None)
    best_validation_metrics_loss: float | None = Field(default=None)
    checkpoint_number: int | None = Field(default=None)
    epoch: int | None = Field(default=None)
    last_improvement_steps: int | None = Field(default=None)
    learning_rate: float | None = Field(default=None)
    steps: int | None = Field(default=None)
    total_tokens_used: int | None = Field(default=None)
    train_metrics_loss: float | None = Field(default=None)
    validation_metrics_loss: float | None = Field(default=None)
    grpo_train_rewards: _GRPOTrainRewards | None = Field(default=None)


class _FinetuningMetricsMeta(BaseModel):
    is_completed: bool
    job_uuid: str = Field(validation_alias="model_uuid")
    run_id: str
    steps: int
    steps_per_epoch: int
    total_steps: int


class _FinetuningMetrics(BaseModel):
    data: _FinetuningMetricsData
    meta: _FinetuningMetricsMeta



import itertools
import mimetypes
import os
import tempfile
from typing import List, Optional, Union

import pandas as pd
import requests
from google.protobuf.json_format import MessageToJson
from predibase_api.artifacts.v1.artifacts_pb2 import (
    ARTIFACT_TYPE_DATASET,
    PresignedUrlForUploadRequest,
    PresignedUrlForUploadResponse,
    RegisterUploadedFileAsDatasetRequest,
)

from predibase.pql import Session
from predibase.pql.api import ServerResponseError
from predibase.resource.connection import get_dataset
from predibase.resource.dataset import Dataset


class DatasetMixin:
    session: Session

    def list_datasets(
        self,
        name: str = None,
        limit: int = 9999999,
        df: bool = False,
    ) -> Union[pd.DataFrame, List[Dataset]]:
        """
        :param name: filter by dataset name
        :param limit: max number of datasets to return
        :param df: Whether to return pandas dataframe or list of objects
        :return: Pandas dataframe or list of datasets
        """
        endpoint = f"/datasets?limit={limit}"
        if name:
            endpoint += "&searchKeys[]=name&searchVals[]=" + name
        resp = self.session.get_json(endpoint)
        datasets = [x for x in resp["datasets"]]
        if df:
            return pd.DataFrame(datasets).drop(columns=["connection"])
        from predibase.resource_util import build_dataset

        return [build_dataset(x, self.session) for x in datasets]

    def list_public_datasets(self) -> Union[pd.DataFrame]:
        """Function to list the public datasets available for import.

        Returns: Pandas dataframe of public datasets
        """

        # Get public datasets
        resp = self.session.get_json("/datasets/get_ludwig_datasets")
        datasets = [x["name"] for x in resp["datasets"]]
        tasks = [", ".join(set(x["tasks"])) for x in resp["datasets"]]

        # Get public datasets connection
        endpoint = "/connections?searchKeys[]=name&searchVals[]=public_datasets"
        resp = self.session.get_json(endpoint)
        connection_id = resp["connections"][0]["id"]

        # Get already imported datasets
        endpoint = f"/datasets?indexVals[]={connection_id}&indexKeys[]=connectionID&limit=100"
        resp = self.session.get_json(endpoint)
        imported_datasets = [True if x in [x["name"] for x in resp["datasets"]] else False for x in datasets]

        return pd.DataFrame({"Name": datasets, "Tasks": tasks, "Imported": imported_datasets})

    def get_dataset(
        self,
        dataset_name: Optional[str] = None,
        connection_name: Optional[str] = None,
        dataset_id: Optional[int] = None,
    ) -> Dataset:
        return get_dataset(self.session, dataset_name, connection_name, dataset_id)

    def delete_dataset(self, dataset_name: str, connection_name: Optional[str] = None):
        dataset = self.get_dataset(dataset_name=dataset_name, connection_name=connection_name)
        return self.session.delete_json(f"/datasets/{dataset.id}")

    def create_dataset_from_df(self, df: pd.DataFrame, name: str) -> Dataset:
        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, f"{name}.parquet")
            df.to_parquet(path)
            return self._upload_file(path, name)

    def upload_file(self, file_path: str, name: str) -> Dataset:
        return self._upload_file(file_path, name)

    def check_file_dataset_exists(self, name: str) -> bool:
        resp = self.session.head(f"/datasets/check_file_dataset_exists/{name}")

        if resp.status_code == 200:
            return False

        if resp.status_code == 400:
            return True

        raise ServerResponseError("Failed to check dataset name", resp.status_code)

    def upload_dataset(self, file_path: str, name: Optional[str] = None) -> Dataset:
        if name is None:
            # Autogenerate a name
            name = os.path.basename(file_path).split(".")[0]
            for i in itertools.count(start=1):
                try:
                    exists = self.check_file_dataset_exists(name)
                except ServerResponseError:
                    exists = False

                if not exists:
                    break

                name = f"{name}_{i}"

            print(f"Dataset name was not explicitly provided. Defaulting to: {name}")

        print("Uploading dataset...")
        res = self._upload_file(file_path, name)
        print("Dataset uploaded.")
        return res

    def _upload_file(self, file_path: str, name: str) -> Dataset:
        if not self.session.is_plan_expired():
            with open(file_path, "rb") as f:
                file_name = os.path.basename(file_path)
                mime_type = mimetypes.guess_type(file_path)[0] or "text/plain"

                # Get presigned url for upload
                presigned_url_req = MessageToJson(
                    PresignedUrlForUploadRequest(
                        mime_type=mime_type,
                        expiry_seconds=3600,
                        artifact_type=ARTIFACT_TYPE_DATASET,
                    ),
                    preserving_proto_field_name=True,
                )
                resp = PresignedUrlForUploadResponse(
                    **self.session.post(
                        "/datasets/get_presigned_url",
                        data=presigned_url_req,
                    ),
                )

                # Unpack response data
                presigned_url = resp.url
                object_name = resp.object_name
                required_headers = resp.required_headers

                # Get required headers for presigned url upload
                headers = {"Content-Type": mime_type}
                for k, v in required_headers:
                    headers[k] = v

                # Upload file to blob storage with pre-signed url
                requests.put(presigned_url, data=f, headers=headers).raise_for_status()

                # Register uploaded file as dataset
                register_dataset_req = MessageToJson(
                    RegisterUploadedFileAsDatasetRequest(
                        dataset_name=name,
                        object_name=object_name,
                        file_name=file_name,
                    ),
                    preserving_proto_field_name=True,
                )
                resp = self.session.post("/datasets/register_uploaded_file", data=register_dataset_req)

                dataset_id = resp["id"]
                # TODO(hungcs): don't wait for dataset but move dataset creation outside workflow to be synchronous
                endpoint = f"/datasets/{dataset_id}"
                resp = self.session.wait_for_dataset(endpoint, until_fully_connected=True)

                from predibase.resource_util import build_dataset

                return build_dataset(resp, self.session)
        else:
            raise PermissionError(
                "Connecting data is locked for expired plans. Contact us to upgrade.",
            )

    def import_public_dataset(self, dataset_name: str) -> Dataset:
        """Function to import a public dataset.

        Args:
            dataset_name: Name of the public dataset to import

        Returns: Dataset
        """
        if not self.session.is_plan_expired():
            public_datasets = self.list_public_datasets()

            if public_datasets[public_datasets["Name"] == dataset_name]["Imported"].values[0]:
                raise Exception(f"Dataset {dataset_name} has already been imported.")

            data = {"selectedDataset": dataset_name}

            resp = self.session.post("/datasets/upload_ludwig_dataset", json=data)
            dataset_id = resp["id"]
            endpoint = f"/datasets/{dataset_id}"
            resp = self.session.wait_for_dataset(endpoint, until_fully_connected=True)

            from predibase.resource_util import build_dataset

            return build_dataset(resp, self.session)
        else:
            raise PermissionError(
                "Connecting data is locked for expired plans. Contact us to upgrade.",
            )



import copy
import os
from dataclasses import dataclass, field
from typing import Optional, Tuple, Union

import yaml
from tabulate import tabulate

from predibase.resource import model as mdl
from predibase.resource.dataset import Dataset
from predibase.resource.llm import interface

_PATH_HERE = os.path.abspath(os.path.dirname(__file__))
_TEMPLATE_DIR = os.path.join(_PATH_HERE, "templates")
_CONFIG_FILENAME = "config.yaml"

LLM_FINETUNE_TEMPLATES_INFO = {}


@dataclass
class FinetuneTemplate:
    llm: "interface.HuggingFaceLLM"
    filename: str
    _meta: Optional[dict] = field(default=None, init=False)
    _raw_template_str: Optional[str] = field(default=None, init=False)
    _cfg: Optional[dict] = field(default=None, init=False)

    @property
    def name(self) -> str:
        if self._meta is None:
            self._load()
        return self._meta["name"]

    @property
    def description(self) -> str:
        if self._meta is None:
            self._load()
        return self._meta["description"]

    def run(
        self,
        prompt_template: str,
        target: str,
        dataset: Optional[Union[str, Dataset]] = None,
        repo: Optional[str] = None,
        epochs: Optional[int] = None,
        train_steps: Optional[int] = None,
        learning_rate: Optional[float] = None,
    ) -> "mdl.ModelFuture":
        return self.llm.finetune(
            dataset=dataset,
            config=self.to_config(prompt_template, target),
            repo=repo,
            epochs=epochs,
            train_steps=train_steps,
            learning_rate=learning_rate,
        )

    def to_config(
        self,
        prompt_template: str = "__PROMPT_TEMPLATE_PLACEHOLDER__",
        target: str = "__TARGET_PLACEHOLDER__",
    ) -> dict:
        if self._cfg is None:
            self._cfg = yaml.safe_load(
                self._raw_template_str.format(
                    base_model=self.llm.model_name,
                    prompt_template=prompt_template,
                    target=target,
                ),
            )
            del self._cfg["template_meta"]

        return copy.deepcopy(self._cfg)

    def _load(self):
        with open(os.path.join(_TEMPLATE_DIR, self.filename)) as f:
            self._raw_template_str = f.read()

        self._meta = yaml.safe_load(self._raw_template_str)["template_meta"]


class FinetuneTemplateCollection(dict):
    def __init__(self, llm: "interface.HuggingFaceLLM"):
        super().__init__()

        self.llm = llm
        for filename in os.listdir(_TEMPLATE_DIR):
            if not filename.endswith(".yaml"):
                continue

            if filename == _CONFIG_FILENAME:
                with open(os.path.join(_TEMPLATE_DIR, _CONFIG_FILENAME)) as config_file:
                    self._config = yaml.safe_load(config_file)
            else:
                tmpl = FinetuneTemplate(llm, os.path.join(_TEMPLATE_DIR, filename))
                self[tmpl.name] = tmpl

    @property
    def default(self) -> FinetuneTemplate:
        default_name = self._config.get("default", "")
        tmpl = self.get(default_name, None)
        if tmpl is None:
            raise RuntimeError(f"the default template '{default_name}' was not found")

        return tmpl

    def compare(self):
        print(self)

    def __str__(self):
        def make_row(key) -> Tuple[str, str, str]:
            tmpl = self[key]
            default_indicator = "------>" if tmpl.name == self._config["default"] else ""
            return default_indicator, key, tmpl.description

        return tabulate(
            (make_row(k) for k in sorted(self.keys())),
            tablefmt="simple_grid",
            headers=["Default", "Name", "Description"],
            colalign=["center"],
        )



def build_dataset(resp, session):
    if "connection" in resp and resp["connection"] and isinstance(resp["connection"], dict):
        from predibase.resource.connection import Connection

        resp["connection"] = Connection.from_dict({"session": session, **resp["connection"]})
    from predibase.resource.dataset import Dataset

    return Dataset.from_dict({"session": session, **resp})


def build_model_repo(resp, session):
    if "dataset" in resp and resp["dataset"] and isinstance(resp["dataset"], dict):
        resp["dataset"] = build_dataset(resp["dataset"], session)
    from predibase.resource.model import ModelRepo

    return ModelRepo.from_dict({"session": session, **resp})


def build_engine(resp, session):
    from predibase.resource.engine import Engine

    return Engine.from_dict({"session": session, **resp})


def build_model(resp, session):
    if "dataset" in resp and resp["dataset"] and isinstance(resp["dataset"], dict):
        resp["dataset"] = build_dataset(resp["dataset"], session)
    if "repo" in resp and resp["repo"] and isinstance(resp["repo"], dict):
        resp["repo"] = build_model_repo(resp["repo"], session)
    if "engine" in resp and resp["engine"] and isinstance(resp["engine"], dict):
        resp["engine"] = build_engine(resp["engine"], session)
    from predibase.resource.model import Model

    return Model.from_dict({"session": session, **resp})



from __future__ import annotations

from typing import TYPE_CHECKING

from predibase.beta.resources.batch_inference_jobs import BatchInferenceJobs

if TYPE_CHECKING:
    from predibase import Predibase


class Beta:
    def __init__(self, client: Predibase):
        self.batch_inference = BatchInferenceJobs(client)



from __future__ import annotations

from typing import TYPE_CHECKING

from predibase.resources.repo import Repo

if TYPE_CHECKING:
    from predibase import Predibase


class Repos:
    def __init__(self, client: Predibase):
        self._client = client

    def create(self, *, name: str, description: str | None = None, exists_ok: bool = False) -> Repo:
        resp = self._client.http_post(
            "/v2/repos",
            json={
                "name": name,
                "description": description,
                "existsOk": exists_ok,
            },
        )
        return Repo.model_validate(resp)

    def get(self, repo_ref: str | Repo) -> Repo:
        if isinstance(repo_ref, Repo):
            repo_ref = repo_ref.name

        return Repo.model_validate(self._client.http_get(f"/v2/repos/{repo_ref}"))

    def list(self, limit: int = 10) -> list[Repo]:
        resp = self._client.http_get(f"/v2/repos?limit={limit}")
        return [Repo.model_validate(r) for r in resp["data"]["repos"]]



import os
import queue
import ssl
from functools import partial

import gevent.ssl as gssl
import numpy as np
import pandas as pd

# importing triton client libraries
import tritonclient.grpc as grpcclient
import tritonclient.grpc.aio as aio_grpcclient
import tritonclient.http as httpclient
import tritonclient.http.aio as aio_httpclient
from tritonclient.utils import InferenceServerException, triton_to_np_dtype

# import predibase predictor classes
from predibase.predictor import FeatureMetadata, ModelMetadata


def get_value(md: FeatureMetadata, row: pd.Series):
    """Returns the value as an array of 1 record, handling missing values."""
    col = md.name
    dt = md.datatype
    v = row[col]
    if dt == "BYTES":
        if row[col] is None:
            v = b""  # Default missing BYTES to empty string
        else:
            v = str(v).encode("utf-8")  # Cast any no-string to string
    elif row[col] is None:
        v = 0  # Default missing numbers to 0
    return np.array([v], dtype=triton_to_np_dtype(dt))


def get_input_data(metadata: ModelMetadata, row: pd.Series):
    """Create a ditionary for a row that matches the triton numpy format, encoding any strings."""
    d = {}
    for md in metadata.inputs:
        d[md.name] = get_value(md, row)
    return d


def get_data_iterator(metadata: ModelMetadata, df: pd.DataFrame):
    """Turns a pandas data frame into a data iterator with request_id=index, and single batch per row."""
    for i, row in df.iterrows():
        yield i, get_input_data(metadata, row)  # Return the index as its original type


def format_key(name):
    return name.replace("::", "_")


def get_output_data(model_outputs, results):
    for request_id, result in results:
        # If we hit an exception raise it immediately
        if isinstance(result, InferenceServerException):
            raise result
        d = {"request_id": request_id}
        for name in model_outputs:
            d[format_key(name)] = result.as_numpy(name)[0]  # Assume single result per record
        yield d


def triton_metadata_http(endpoint, model_name, headers, model_version=""):
    with httpclient.InferenceServerClient(
        endpoint,
        ssl=True,
        ssl_context_factory=gssl.create_default_context,
    ) as triton_client:
        return triton_client.get_model_metadata(model_name, model_version=model_version, headers=headers)


async def triton_metadata_http_async(endpoint, model_name, headers, model_version=""):
    async with aio_httpclient.InferenceServerClient(  # pylint: disable=unexpected-keyword-arg
        endpoint,
        ssl=True,
        ssl_context=ssl.create_default_context(),
    ) as triton_client:
        return await triton_client.get_model_metadata(model_name, model_version=model_version, headers=headers)


def triton_predict_http(
    endpoint,
    model_name,
    model_inputs,
    model_outputs,
    binary_data,
    headers,
    request_id,
    data_to_predict,
    model_version="",
):
    with httpclient.InferenceServerClient(
        endpoint,
        ssl=True,
        ssl_context_factory=gssl.create_default_context,
    ) as triton_client:  # pylint: disable=unexpected-keyword-arg
        # Define the http inputs
        inputs = [httpclient.InferInput(*params) for params in model_inputs]

        # Set input values
        for i, args in enumerate(model_inputs):
            d = data_to_predict[args[0]]
            inputs[i].set_data_from_numpy(d, binary_data=binary_data)

        # Define the http outputs
        outputs = [httpclient.InferRequestedOutput(name, binary_data=binary_data) for name in model_outputs]

        # Make inference
        return triton_client.infer(
            model_name,
            inputs,
            model_version=model_version,
            outputs=outputs,
            headers=headers,
            request_id=str(request_id),
        )


async def triton_predict_http_async(
    endpoint,
    model_name,
    model_inputs,
    model_outputs,
    binary_data,
    headers,
    request_id,
    data_to_predict,
    model_version="",
):
    async with aio_httpclient.InferenceServerClient(  # pylint: disable=unexpected-keyword-arg
        endpoint,
        ssl=True,
        ssl_context=ssl.create_default_context(),
    ) as triton_client:
        # Define the http inputs
        inputs = [aio_httpclient.InferInput(*params) for params in model_inputs]

        # Set input values
        for i, args in enumerate(model_inputs):
            d = data_to_predict[args[0]]
            inputs[i].set_data_from_numpy(d, binary_data=binary_data)

        # Define the http outputs
        outputs = [aio_httpclient.InferRequestedOutput(name, binary_data=binary_data) for name in model_outputs]

        # Make inference
        return await triton_client.infer(
            model_name,
            inputs,
            model_version=model_version,
            outputs=outputs,
            headers=headers,
            request_id=str(request_id),
        )


def triton_predict_grpc(
    endpoint,
    model_name,
    model_inputs,
    model_outputs,
    creds,
    headers,
    request_id,
    data_to_predict,
    model_version="",
):
    with grpcclient.InferenceServerClient(endpoint, creds=creds) as triton_client:
        # Define the grpc inputs
        inputs = [grpcclient.InferInput(*params) for params in model_inputs]

        # Set input values
        for i, args in enumerate(model_inputs):
            inputs[i].set_data_from_numpy(data_to_predict[args[0]])

        # Define the grpc outputs
        outputs = [grpcclient.InferRequestedOutput(name) for name in model_outputs]

        # Make inference
        return triton_client.infer(
            model_name,
            inputs,
            model_version=model_version,
            outputs=outputs,
            headers=headers,
            request_id=str(request_id),
        )


async def triton_predict_grpc_async(
    endpoint,
    model_name,
    model_inputs,
    model_outputs,
    creds,
    headers,
    request_id,
    data_to_predict,
    model_version="",
):
    async with aio_grpcclient.InferenceServerClient(endpoint, creds=creds) as triton_client:
        # Define the grpc inputs
        inputs = [aio_grpcclient.InferInput(*params) for params in model_inputs]

        # Set input values
        for i, args in enumerate(model_inputs):
            inputs[i].set_data_from_numpy(data_to_predict[args[0]])

        # Define the grpc outputs
        outputs = [aio_grpcclient.InferRequestedOutput(name) for name in model_outputs]

        # Make inference
        return await triton_client.infer(
            model_name,
            inputs,
            model_version=model_version,
            outputs=outputs,
            headers=headers,
            request_id=str(request_id),
        )


class UserData:
    def __init__(self):
        self._completed_requests = queue.Queue()


def callback(user_data, result, error):
    if error:
        user_data._completed_requests.put(error)
    else:
        user_data._completed_requests.put(result)


def get_grpc_creds(token: str):
    import grpc

    secure_cacert: str = os.getenv("CURL_CA_BUNDLE")
    if secure_cacert is not None:
        with open(secure_cacert, "rb") as fd:
            root_c = fd.read()
        scc = grpc.ssl_channel_credentials(root_certificates=root_c)
    else:
        scc = grpc.ssl_channel_credentials()
    tok = grpc.access_token_call_credentials(token)
    return grpc.composite_channel_credentials(scc, tok)


def triton_predict_grpc_stream(endpoint, model_name, model_inputs, creds, headers, data_iterator, model_version=""):
    """This function will iterate of the data iterator which is a tuple of (request_id, data_to_predict)"""
    user_data = UserData()

    with grpcclient.InferenceServerClient(endpoint, creds=creds) as triton_client:
        # Start stream registering callback
        triton_client.start_stream(callback=partial(callback, user_data), headers=headers)

        request_ids = []
        for request_id, data_to_predict in data_iterator:
            # Define the grpc inputs
            inputs = [grpcclient.InferInput(*params) for params in model_inputs]

            # Set input values
            for i, args in enumerate(model_inputs):
                inputs[i].set_data_from_numpy(data_to_predict[args[0]])

            # Make inference
            request_ids.append(request_id)
            triton_client.async_stream_infer(
                model_name=model_name,
                inputs=inputs,
                model_version=model_version,
                request_id=str(request_id),
            )

        # Return all results
        for request_id in request_ids:
            try:
                result = user_data._completed_requests.get()
                yield request_id, result
            except queue.Empty:
                break


async def async_stream_yield(data_iterator, model_name, model_version, model_inputs):
    for request_id, data_to_predict in data_iterator:
        # Define the grpc inputs
        inputs = [aio_grpcclient.InferInput(*params) for params in model_inputs]

        # Set input values
        for i, args in enumerate(model_inputs):
            inputs[i].set_data_from_numpy(data_to_predict[args[0]])

        # Issue the asynchronous sequence inference.
        yield {
            "model_name": model_name,
            "model_version": model_version,
            "inputs": inputs,
            "request_id": str(request_id),
        }


async def triton_predict_grpc_stream_async(
    endpoint,
    model_name,
    model_inputs,
    creds,
    headers,
    data_iterator,
    model_version="",
):
    """This function will iterate of the data iterator which is a tuple of (request_id, data_to_predict)"""
    async with aio_grpcclient.InferenceServerClient(endpoint, creds=creds) as triton_client:
        # Request iterator that yields the next request
        async def async_request_iterator():
            async for request in async_stream_yield(data_iterator, model_name, model_version, model_inputs):
                yield request

        # Start streaming
        response_iterator = triton_client.stream_infer(
            inputs_iterator=async_request_iterator(),
            headers=headers,
        )

        # Read response from the stream
        async for response in response_iterator:
            result, error = response
            if error:
                yield None, error
            else:
                request_id = result.get_response().id.split("_")[0]
                yield request_id, result



from pydantic import TypeAdapter


class ValidatedDict(dict):
    def __init__(self, kt: type, vt: type, *args, **kwargs):
        self._kt = kt
        self._vt = vt
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        key = TypeAdapter(self._kt).validate_python(key)
        value = TypeAdapter(self._vt).validate_python(value)
        super().__setitem__(key, value)





from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import AliasPath, BaseModel, Field

from predibase.config import UpdateDeploymentConfig

if TYPE_CHECKING:
    pass


class Deployment(BaseModel):
    name: str
    uuid: str
    description: str | None
    type: str
    status: str
    context_window: int = Field(validation_alias=AliasPath("model", "maxInputLength"))
    accelerator: str = Field(validation_alias=AliasPath("accelerator", "id"))
    quantization: str | None = Field(default=None)
    model: str = Field(validation_alias=AliasPath("model", "name"))
    current_replicas: int = Field(validation_alias="currentReplicas")
    config: UpdateDeploymentConfig = Field(
        validation_alias="config",
    )
    uses_guaranteed_capacity: bool = Field(
        validation_alias="usesGuaranteedCapacity",
    )



import inspect
import logging
import time
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd


def get_results_df(data: dict):
    dataset = data["dataset"]
    if "data" not in dataset:
        return pd.DataFrame()

    return pd.DataFrame(
        data=dataset["data"],
        columns=dataset["columns"],
        index=dataset.get("index"),
    )


def remove_trailing_slash(url):
    while url.endswith("/"):
        url = url[:-1]
    return url


# from: https://github.com/fugue-project/triad/blob/master/triad/utils/convert.py
def get_caller_global_local_vars(
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
    start: int = -1,
    end: int = -1,
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Get the caller level global and local variables.

    :param global_vars: overriding global variables, if not None,
      will return this instead of the caller's globals(), defaults to None
    :param local_vars: overriding local variables, if not None,
      will return this instead of the caller's locals(), defaults to None
    :param start: start stack level (from 0 to any negative number),
      defaults to -1 which is one level above where this function is invoked
    :param end: end stack level (from ``start`` to any smaller negative number),
      defaults to -1 which is one level above where this function is invoked
    :return: tuple of `global_vars` and `local_vars`
    :Examples:
        .. code-block:: python
            def caller():
                x = 1
                assert 1 == get_value("x")


            def get_value(var_name):
                _, l = get_caller_global_local_vars()
                assert var_name in l
                assert var_name not in locals()
                return l[var_name]
    :Notice:
    This is for internal use, users normally should not call this directly.
    If merging multiple levels, the variables on closer level
    (to where it is invoked) will overwrite the further levels values if there
    is overlap.
    :Examples:
        .. code-block:: python
            def f1():
                x = 1

                def f2():
                    x = 2

                    def f3():
                        _, l = get_caller_global_local_vars(start=-1, end=-2)
                        assert 2 == l["x"]
                        _, l = get_caller_global_local_vars(start=-2, end=-2)
                        assert 1 == l["x"]

                f2()


            f1()
    """
    stack = inspect.currentframe().f_back  # type: ignore
    p = 0
    while p > start and stack is not None:
        stack = stack.f_back
        p -= 1
    g_arr: List[Dict[str, Any]] = []
    l_arr: List[Dict[str, Any]] = []
    while p >= end and stack is not None:
        g_arr.insert(0, stack.f_globals)
        l_arr.insert(0, stack.f_locals)
        stack = stack.f_back
        p -= 1
    if global_vars is None:
        global_vars = {}
        for d in g_arr:
            global_vars.update(d)
    if local_vars is None:
        local_vars = {}
        for d in l_arr:
            local_vars.update(d)
    return global_vars, local_vars  # type: ignore


def retry(times, exceptions):
    """Retry Decorator Retries the wrapped function/method `times` times if the exceptions listed in ``exceptions``
    are thrown.

    Source: https://stackoverflow.com/a/64030200

    :param times: The number of times to repeat the wrapped function/method
    :type times: int
    :param Exceptions: Lists of exceptions that trigger a retry attempt
    :type Exceptions: Tuple[Exception]
    """

    def decorator(func):
        def newfn(*args, **kwargs):
            attempt = 0
            while attempt < times:
                try:
                    return func(*args, **kwargs)
                except exceptions:
                    logging.warning(
                        "Exception thrown when attempting to run %s, attempt " "%d of %d" % (func, attempt, times),
                    )
                    attempt += 1
                    time.sleep(1.0)
            return func(*args, **kwargs)

        return newfn

    return decorator



from __future__ import annotations

import ast
import base64
import inspect
from functools import cached_property
from typing import Callable, Literal, TYPE_CHECKING, Any

from predibase.resources.validation import ValidatedDict
from pydantic import BaseModel, ConfigDict, Field, field_validator, PositiveFloat, PositiveInt, model_validator, \
    AliasChoices

if TYPE_CHECKING:
    pass


class SamplingParamsConfig(BaseModel):
    temperature: float | None = Field(default=None)
    top_p: float | None = Field(default=None)
    top_k: int | None = Field(default=None)
    max_tokens: int | None = Field(default=None)


class FinetuningConfig(BaseModel):
    # This is required in most cases, but coded as optional for continued training.
    base_model: str | None = Field(default=None)

    adapter: str | None = Field(default=None)
    task: Literal["sft", "continued_pretraining", "grpo"] | None = Field(default=None)
    epochs: int | None = Field(default=None)
    learning_rate: float | None = Field(default=None)
    rank: int | None = Field(default=None)
    target_modules: list[str] | None = Field(default=None)
    enable_early_stopping: bool | None = Field(default=None)
    apply_chat_template: bool | None = Field(default=None)
    lr_scheduler: dict | None = Field(default=None)
    optimizer: dict | None = Field(default=None)
    lora_alpha: PositiveInt | None = Field(default=None)
    lora_dropout: PositiveFloat | None = Field(default=None)
    warmup_ratio: PositiveFloat | None = Field(default=None)
    effective_batch_size: PositiveInt | None = Field(default=None)


class SFTConfig(FinetuningConfig):
    task: Literal["sft"] = "sft"


class ContinuedPretrainingConfig(FinetuningConfig):
    task: Literal["continued_pretraining"] = "continued_pretraining"


class GRPOConfig(FinetuningConfig):
    task: Literal["grpo"] = "grpo"
    beta: float | None = Field(default=None)
    num_generations: int | None = Field(default=None)
    sampling_params: SamplingParamsConfig | None = Field(default=None)
    reward_fns: RewardFunctionsConfig | None = Field(default=None,
                                                     serialization_alias="rewardFns",
                                                     validation_alias=AliasChoices("rewardFns", "reward_fns"))

    @field_validator("reward_fns", mode="before")
    @classmethod
    def convert_callable_list(cls, val: Any) -> Any:
        if not isinstance(val, list):
            return val

        return RewardFunctionsConfig(functions=val)


def get_task_type(cfg: FinetuningConfig, default: str = "sft") -> str:
    if isinstance(cfg, dict):
        return str(cfg.get("task", default))

    return cfg.task


class RewardFunctionsConfig(BaseModel, validate_assignment=True):
    runtime: RewardFunctionsRuntimeConfig | None = Field(default=None)
    functions: dict[str, RewardFunction] = Field(default_factory=dict)

    @field_validator("functions", mode="before")
    @classmethod
    def convert_callable_list(cls, val: Any) -> Any:
        if not isinstance(val, list):
            return val

        return {f.__name__: RewardFunction.from_callable(f) for f in val}

    @field_validator("functions", mode="after")
    @classmethod
    def convert_to_validated_dict(cls, val: dict[str, RewardFunction]) -> dict[str, RewardFunction]:
        return ValidatedDict(str, RewardFunction, val)

    def __setitem__(self, key: str, value: Callable | RewardFunction):
        if not isinstance(key, str):
            raise ValueError(f"expected key to be a string - got {type(key)}")

        if not isinstance(value, Callable) and not isinstance(value, RewardFunction):
            raise ValueError(f"expected value to be a callable or RewardFunction object - got {type(value)}")

        self.functions[key] = value

    def __getitem__(self, key: str):
        return self.functions[key]

    def __delitem__(self, key: str):
        return self.functions.__delitem__(key)


class RewardFunctionsRuntimeConfig(BaseModel):
    packages: list[str] | None = Field(default=None)


class RewardFunction(BaseModel):
    encoded_fn: str = Field(serialization_alias="encodedFn", validation_alias=AliasChoices("encodedFn", "encoded_fn"))

    def __repr__(self):
        encoded_fn_snippet = f"{self.encoded_fn[:30]}..." if len(self.encoded_fn) >= 30 else self.encoded_fn
        return f"RewardFunction(encoded_fn='{encoded_fn_snippet}')"

    @model_validator(mode="before")
    @classmethod
    def coerce_from_callable(cls, val: Any) -> Any:
        if isinstance(val, Callable):
            # NOTE: the call to model_dump is necessary - see
            # https://github.com/pydantic/pydantic/issues/9336#issuecomment-2082419008
            val = RewardFunction.from_callable(val).model_dump()

        return val

    @classmethod
    def from_callable(cls, fn: Callable) -> RewardFunction:
        fn_name = getattr(fn, "__name__", None)
        if not fn_name:
            raise ValueError(f"Input is not a function or is missing a name")

        if not inspect.isfunction(fn):
            raise ValueError(f"Expected input to be a function")

        params = inspect.signature(fn).parameters
        if len(params) != 3:
            raise ValueError(f"unexpected number of arguments for input - expected 3, "
                             f"found {len(params)}")

        fn_source = inspect.getsource(fn)
        return RewardFunction(encoded_fn=base64.b64encode(fn_source.encode("utf-8")).decode("utf-8"))

    @cached_property
    def source(self):
        return base64.b64decode(self.encoded_fn).decode("utf-8")

    @cached_property
    def function(self):
        ast_root = ast.parse(self.source)

        # Check that the input is a single function definition
        if len(ast_root.body) != 1 or not isinstance(ast_root.body[0], ast.FunctionDef):
            raise ValueError("Expected encoded data to be a Python function")

        # Render the function locally
        exec(compile(ast_root, filename="<string>", mode="exec"))

        return locals()[ast_root.body[0].name]


class UpdateDeploymentConfig(BaseModel):
    # Note: necessary because this particular model and its children are used to both deserialize and serialize data.
    # Other models in this SDK are used exclusively for deserializing data to a Python object that is never (for now)
    # reserialized back to the API for a different request.
    model_config = ConfigDict(populate_by_name=True)

    custom_args: list[str] | None = Field(
        default=None,
        validation_alias="customArgs",
        serialization_alias="customArgs",
    )
    cooldown_time: int | None = Field(
        default=None,
        validation_alias="cooldownTime",
        serialization_alias="cooldownTime",
    )
    hf_token: str | None = Field(
        default=None,
        validation_alias="hfToken",
        serialization_alias="hfToken",
    )
    min_replicas: int | None = Field(
        default=None,
        validation_alias="minReplicas",
        serialization_alias="minReplicas",
    )
    max_replicas: int | None = Field(
        default=None,
        validation_alias="maxReplicas",
        serialization_alias="maxReplicas",
    )
    scale_up_threshold: int | None = Field(
        default=None,
        validation_alias="scaleUpRequestThreshold",
        serialization_alias="scaleUpRequestThreshold",
    )
    max_total_tokens: int | None = Field(
        default=None,
        validation_alias="maxTotalTokens",
        serialization_alias="maxTotalTokens",
    )
    lorax_image_tag: str | None = Field(
        default=None,
        validation_alias="loraxImageTag",
        serialization_alias="loraxImageTag",
    )
    request_logging_enabled: bool | None = Field(
        default=None,
        validation_alias="requestLoggingEnabled",
        serialization_alias="requestLoggingEnabled",
    )
    direct_ingress: bool | None = Field(
        default=None,
        validation_alias="directIngress",
        serialization_alias="directIngress",
    )
    preloaded_adapters: list[str] | None = Field(
        default=None,
        validation_alias="preloadedAdapters",
        serialization_alias="preloadedAdapters",
    )
    speculator: str | None = Field(
        default=None,
        validation_alias="speculator",
        serialization_alias="speculator",
    )
    prefix_caching: bool | None = Field(
        default=None,
        validation_alias="prefixCaching",
        serialization_alias="prefixCaching",
    )

    backend: Literal["v1", "v2"] | None = Field(
        default=None,
        validation_alias="backend",
        serialization_alias="backend"
    )


class DeploymentConfig(UpdateDeploymentConfig):
    base_model: str = Field(
        ...,
        validation_alias="baseModel",
        serialization_alias="baseModel",
    )
    accelerator: str | None = Field(default=None)
    uses_guaranteed_capacity: bool | None = Field(
        default=None,
        validation_alias="usesGuaranteedCapacity",
        serialization_alias="usesGuaranteedCapacity",
    )
    quantization: str | None = Field(default=None)


class AugmentationConfig(BaseModel):
    """Configuration for synthetic data generation tasks.

    # Attributes
    :param base_model: (str) The OpenAI model to prompt.
    :param num_samples_to_generate: (int) The number of synthetic examples to generate.
    :param num_seed_samples: (int) The number of seed samples to use for generating synthetic examples.
    :param task_context: (str) The user-provided task context for generating candidates.
    """

    base_model: str
    num_samples_to_generate: int = Field(default=1000)
    num_seed_samples: int | str = Field(default="all")
    augmentation_strategy: str = Field(default="mixture_of_agents")
    task_context: str = Field(default="")

    @field_validator("base_model")
    @classmethod
    def validate_base_model(cls, base_model) -> str:
        supoorted_base_models = {
            "gpt-4-turbo",
            "gpt-4-0125-preview",
            "gpt-4-1106-preview",
            "gpt-4o",
            "gpt-4o-2024-08-06",
            "gpt-4o-mini",
        }
        if base_model not in supoorted_base_models:
            raise ValueError(
                f"base_model must be one of {supoorted_base_models}.",
            )
        return base_model

    @field_validator("num_samples_to_generate")
    @classmethod
    def validate_num_samples_to_generate(cls, num_samples_to_generate) -> int:
        if num_samples_to_generate < 1:
            raise ValueError("num_samples_to_generate must be >= 1.")
        return num_samples_to_generate

    @field_validator("num_seed_samples")
    @classmethod
    def validate_num_seed_samples(cls, num_seed_samples):
        if isinstance(num_seed_samples, str) and num_seed_samples != "all":
            raise ValueError("num_seed_samples can only be an integer or the string 'all'.")
        elif isinstance(num_seed_samples, int) and num_seed_samples < 1:
            raise ValueError("num_seed_samples must be >= 1.")
        return num_seed_samples

    @field_validator("augmentation_strategy")
    @classmethod
    def validate_augmentation_strategy(cls, augmentation_strategy):
        if augmentation_strategy not in {"single_pass", "mixture_of_agents"}:
            raise ValueError("augmentation_strategy must be 'single_pass' or 'mixture_of_agents'.")
        return augmentation_strategy



from dataclasses import dataclass
from typing import List

import pandas as pd

from predibase.pql.api import Session


@dataclass
class FeatureMetadata:
    name: str
    datatype: str
    shape: List[int]


@dataclass
class ModelMetadata:
    name: str
    versions: List[str]
    platform: str
    inputs: List[FeatureMetadata]
    outputs: List[FeatureMetadata]

    @classmethod
    def from_dict(self, d):
        return ModelMetadata(
            name=d["name"],
            versions=d["versions"],
            platform=d["platform"],
            inputs=[FeatureMetadata(**md) for md in d["inputs"]],
            outputs=[FeatureMetadata(**md) for md in d["outputs"]],
        )


class Predictions:
    def __init__(self, model_outputs: List[FeatureMetadata], results: List):
        self.model_outputs = model_outputs
        self.results = results

    def to_pandas(
        self,
    ):
        """Convert the output data into pandas data frame."""
        from predibase.triton_util import get_output_data

        data_iterator = get_output_data(self.model_outputs, self.results)
        pred_df = pd.DataFrame(data_iterator).set_index("request_id")
        pred_df.index.rename(None, inplace=True)  # Clear the index name
        return pred_df


class Predictor:
    def __init__(
        self,
        session: Session,
        deployment_name: str,
        deployment_version: int = 0,
    ):
        # Set properties from constructor
        self.session = session
        self.deployment_name = deployment_name
        if deployment_version > 0:
            self.deployment_version = str(deployment_version)
        else:
            self.deployment_version = ""  # Empty string to get latest version
        self.grpc_endpoint = session.serving_grpc_endpoint
        self.http_endpoint = f"{self.session.serving_http_endpoint}/{session.tenant}/deployments"
        self.http_headers = {"Authorization": f"Bearer {self.session.token}"}

    def metadata(self) -> ModelMetadata:
        """Returns the model metadata for a given deployment name and optional version."""
        from predibase.triton_util import triton_metadata_http

        metadata = triton_metadata_http(
            endpoint=self.http_endpoint,
            model_name=self.deployment_name,
            model_version=self.deployment_version,
            headers=self.http_headers,
        )
        return ModelMetadata.from_dict(metadata)

    def predict(self, input_df: pd.DataFrame, stream=False) -> Predictions:
        """Returns predictions for an input dataframe.

        Set stream to true, to fetch results via secure grpc connection.
        """
        try:
            from predibase.triton_util import (
                get_data_iterator,
                get_grpc_creds,
                InferenceServerException,
                triton_predict_grpc_stream,
                triton_predict_http,
            )
        except ModuleNotFoundError:
            print("Predict not supported, please ensure you have installed predictor requirements:")
            print('pip install "predibase[predictor]"')
            return

        try:
            metadata = self.metadata()
        except InferenceServerException as e:
            print(f"Failed to get metadata: {e.message}")
            return

        # Get model inputs and outputs from metadata
        model_inputs = [(md.name, md.shape, md.datatype) for md in metadata.inputs]
        model_outputs = [md.name for md in metadata.outputs]

        # Get the data iterator using model metadata
        data_iterator = get_data_iterator(metadata, input_df)

        if stream:
            print(f"predict via gRPC stream: {self.grpc_endpoint}")
            try:
                # Use model name and version to query specific model
                model_name = metadata.name
                model_version = metadata.versions[0]
                grpc_headers = {
                    "tenant": self.session.tenant,
                    "model_name": model_name,
                    "model_version": str(model_version),
                }
                # Get grpc credentials
                grpc_creds = get_grpc_creds(self.session.token)
                results = triton_predict_grpc_stream(
                    endpoint=self.grpc_endpoint,
                    headers=grpc_headers,
                    creds=grpc_creds,
                    model_name=model_name,
                    model_version=model_version,
                    model_inputs=model_inputs,
                    data_iterator=data_iterator,
                )
            except Exception as e:
                print(f"error with stream predict -- {e}")
                stream = False  # fallback to http

        if not stream:
            print(f"predict via REST: {self.http_endpoint}")
            results = []
            for request_id, data in data_iterator:
                result = triton_predict_http(
                    endpoint=self.http_endpoint,
                    model_name=self.deployment_name,
                    model_version=str(self.deployment_version),
                    model_inputs=model_inputs,
                    model_outputs=model_outputs,
                    binary_data=False,
                    request_id=str(request_id),
                    headers=self.http_headers,
                    data_to_predict=data,
                )
                results.append((request_id, result))

        # Return a predictions object, which can lazyily load the predicions
        return Predictions(model_outputs, results)


class AsyncPredictor(Predictor):
    def __init__(
        self,
        session: Session,
        deployment_name: str,
        deployment_version: int = 0,
    ):
        super().__init__(session, deployment_name, deployment_version)

    async def metadata(self) -> ModelMetadata:
        """Returns the model metadata for a given deployment name and optional version."""
        from predibase.triton_util import triton_metadata_http_async

        metadata = await triton_metadata_http_async(
            endpoint=self.http_endpoint,
            model_name=self.deployment_name,
            model_version=self.deployment_version,
            headers=self.http_headers,
        )
        return ModelMetadata.from_dict(metadata)

    async def predict(self, input_df: pd.DataFrame, stream=False) -> Predictions:
        """Returns predictions for an input dataframe.

        Set stream to true, to fetch results via secure grpc connection.
        """
        try:
            from predibase.triton_util import (
                get_data_iterator,
                get_grpc_creds,
                triton_predict_grpc_stream_async,
                triton_predict_http_async,
            )
        except ModuleNotFoundError:
            print("Predict not supported, please ensure you have installed predictor requirements:")
            print('pip install "predibase[predictor]"')
            return

        metadata = await self.metadata()

        # Get model inputs and outputs from metadata
        model_inputs = [(md.name, md.shape, md.datatype) for md in metadata.inputs]
        model_outputs = [md.name for md in metadata.outputs]

        # Get the data iterator using model metadata
        data_iterator = get_data_iterator(metadata, input_df)

        if stream:
            print(f"predict via gRPC stream: {self.grpc_endpoint}")
            try:
                # Use model name and version to query specific model
                model_name = metadata.name
                model_version = metadata.versions[0]
                grpc_headers = {
                    "tenant": self.session.tenant,
                    "model_name": model_name,
                    "model_version": str(model_version),
                }
                # Get grpc credentials
                grpc_creds = get_grpc_creds(self.session.token)
                results = [
                    r
                    async for r in triton_predict_grpc_stream_async(
                        endpoint=self.grpc_endpoint,
                        headers=grpc_headers,
                        creds=grpc_creds,
                        model_name=model_name,
                        model_version=model_version,
                        model_inputs=model_inputs,
                        data_iterator=data_iterator,
                    )
                ]
            except Exception as e:
                print(f"error with stream predict -- {e}")
                stream = False  # fallback to http

        if not stream:
            print(f"predict via REST: {self.http_endpoint}")
            results = []
            for request_id, data in data_iterator:
                result = await triton_predict_http_async(
                    endpoint=self.http_endpoint,
                    model_name=self.deployment_name,
                    model_version=str(self.deployment_version),
                    model_inputs=model_inputs,
                    model_outputs=model_outputs,
                    binary_data=False,
                    request_id=str(request_id),
                    headers=self.http_headers,
                    data_to_predict=data,
                )
                results.append((request_id, result))

        # Return a predictions object, which can lazyily load the predicions
        return Predictions(model_outputs, results)



import json
import os
from pathlib import Path
from typing import Any, Dict

import typer

APP_NAME = "predibase-sdk-app"
LOCAL_SETTINGS_NAME = ".predibase.json"


def get_global_settings_path() -> Path:
    app_dir = typer.get_app_dir(APP_NAME)
    return Path(app_dir) / "settings.json"


def save_global_settings(settings: Dict[str, Any]):
    settings_path = get_global_settings_path()
    os.makedirs(settings_path.parent, exist_ok=True)
    with open(settings_path, "w") as f:
        json.dump(settings, f)


def load_global_settings() -> Dict[str, Any]:
    settings_path = get_global_settings_path()
    if settings_path.is_file():
        with open(settings_path) as f:
            return json.load(f)
    return {}


def save_local_settings(settings: Dict[str, Any]):
    settings_path = Path(LOCAL_SETTINGS_NAME)
    os.makedirs(settings_path.parent, exist_ok=True)
    with open(settings_path, "w") as f:
        json.dump(settings, f)


def load_local_settings() -> Dict[str, Any]:
    settings_path = Path(LOCAL_SETTINGS_NAME)
    if settings_path.is_file():
        with open(settings_path) as f:
            return json.load(f)
    return {}


def load_settings() -> Dict[str, Any]:
    local_settings = load_local_settings()
    global_settings = load_global_settings()
    return {**global_settings, **local_settings}
